/*
 * @author Lenidas
 *
 * Sistema para construo grfica de grafos
 *
 * rea de desenho: desenha quadrados representando os vrtices e linhas para arestas
 *
 * Estrutura:
 *
 * Retangulos: 0..MAXRETANGULOS
 *             java.awt.Point
 * Arestas   : 0..MAXARESTAS
 *  
 * Point     : int x, int y
 *
 * matAdj[][]: matriz de adjacncias
 *             (i)----(j) <=> matAdj[i,j] = matAdj[j,i] = n, nm. aresta que liga os ns (grafo no dirigido)
 */



import java.awt.Checkbox;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Panel;
import java.awt.Point;
import java.awt.event.*;

public class AreaDesenho extends Panel implements MouseMotionListener {

 static final int CROSSHAIR_CURSOR = 1; // setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
                                        // Cursor.HAND_CURSOR));

 static final int RAIO  = 10; // raio;
 static final int steps = 4;
 static final int DELAY = 10;
 
 static final int RETANGULO = 0;
 static final int ARESTA   = 1;
 
 static final String S_RETANGULO = "Retngulo";
 static final String S_ARESTA    = "Aresta";
 
 Image offscreen;
 Graphics gr;
 Color corRetangulo = new Color(0, 100, 155),
       corAresta = new Color(100, 50, 100),
       corBorda = Color.black;
 
 Inicia inicia;

 private static final int LARGURA_RETANGULO = 15; // para largura de se desenhar os retngulos
 private static final int ALTURA_RETANGULO  = 10; // para altura  de se desenhar os retngulos
 private static final int MAXRETANGULOS = 10;
 private Point[] Retangulos  = new Point[MAXRETANGULOS];  // cantos superiores esquerdo dos quadrados
 private Arestas[] arestas  = new Arestas[MAXRETANGULOS]; // se grafo completo

 // Rtulos numricos
 private int[] rotRetangulos = new int[MAXRETANGULOS];    // cantos superiores esquerdo dos quadrados

 // matriz de adjacncias: (i)----(j) => matAdj[i,j]=matAdj[j,i]=n, nm. aresta (grafo no dirigido)
 //         <- quadrados ->|
 //        |               :           |
 // quad.  |               :           |
 //        |...............:...........|
 //        |               :           |
 private int[][] matAdj  = new int[MAXRETANGULOS][MAXRETANGULOS]; 

 private int current = -1;
 private int nquadrados = 0; // nm. atual de quadrados
 private int narestas   = 0; // nm. atual de arestas

 private int nrotquadrados = 0; // nm. do rtulo (s decresce se limpar tudo)

 boolean aresta_prim_clique = true,
         moveObjeto         = false; // qdo clica sobre vrtice ele fica marcado p/ mover

 int moveObjetoI, moveObjetoT;       // ndice e tipo do objeto a ser movido

 // auxiliares para desenhar arestas
 private  int a_i=0, a_f, // ndices (dos vrtices) inicial e final
              t_i=0, t_f; // tipos (dos vrtices) inicial e final



 public AreaDesenho(Inicia inicia) {
   this.addMouseListener(new MouseAdapter() {
       public void mouseClicked(MouseEvent mouseevent) {
   	AreaDesenho.this.mouseClicked(mouseevent);
       }
   });
   this.addMouseMotionListener(this);
   this.inicia = inicia;
   for (int i = 0; i < MAXRETANGULOS; i++)
       for (int j = 0; j <  MAXRETANGULOS; j++)
           matAdj[i][j] = -1;
   }
 
 
 public void Limpa() {
   for (int i = nquadrados - 1; i >= 0; i--)
       removeRetangulo(i);
   for (int i = narestas - 1; i >= 0; i--)
       removeAresta(i);
   for (int i = 0; i < MAXRETANGULOS; i++) {
       for (int j = 0; j <MAXRETANGULOS; j++)
           matAdj[i][j] = matAdj[j][i] = -1;
       }
   nrotquadrados = 0;
   }
 
 // auxiliar p/ verificar matriz de adjacncias (CUIDADO: MAXRETANGULOS grandes => LENTIDO !!!!
 public void imprimeMatriz() {
   int i,j;
   String s="";
   //System.out.println(". MAXRETANGULOS="+MAXRETANGULOS);
   for (i = 0; i < MAXRETANGULOS; i++) {
       for (j = 0; j < MAXRETANGULOS; j++) {
           if ( matAdj[i][j]==-1)
             s = new String(s+" -1");
          else
             s = new String(s+"  "+matAdj[i][j]);
          }
       //System.out.println(i+" "+j);
       s = new String(s+"\n");
       }
   System.out.println(s);
   }

 public void mouseClicked(MouseEvent evt) {
  // pega posio do clique, para (x,y)
   int x = evt.getX(); 
   int y = evt.getY();
   
   int modifiers = evt.getModifiers(); // tipo de evento
   int clickCount = evt.getClickCount();
  
   //System.out.println(modifiers+" "+clickCount);

   if ((modifiers & InputEvent.CTRL_MASK) != 0 && (modifiers & InputEvent.SHIFT_MASK) != 0
  	  //&& clickCount >= 3
      ){
      Graphics g = offscreen.getGraphics(); // g.getGraphics();
      g.translate(getInsets().left, getInsets().top);
      g.drawString("Opss", x, y);
      System.out.println("Opss"+ x + "  "+ y);
      return;
      }
  
   int [] resp = new int[2];
   int tipo;

   resp = find(x, y);
   current = resp[0];

   Checkbox ck = inicia.ck_objeto.getSelectedCheckbox(); //getCurrent();
   if (ck.getLabel().equals(S_RETANGULO))
       tipo = RETANGULO;
   else
   if (ck.getLabel().equals(S_ARESTA))
       tipo = ARESTA;
   else {
       System.out.println("[AreaDesenho!mouseClicked]: erro, objeto inexistente");
       return;
       }

   if (tipo == ARESTA && resp[1]>-1) {
      if (aresta_prim_clique) { // incio da aresta
          a_i = resp[0];
          t_i = resp[1];
          aresta_prim_clique = false;
          inicia.msg("Clique no vrtice final da aresta (aresta de nmero "+narestas+")");
          //System.out.println("-inicio- i="+a_i+" ti="+t_i);
          }
      else {
          insereAresta( a_i, resp[0], t_i, resp[1] );
          aresta_prim_clique = true;
          inicia.msg("Clique no primeiro vrtice para criar aresta                                ");
          //System.out.println("-final- f="+resp[0]+" tf="+resp[1]);
          }
      }
   else aresta_prim_clique = true;

   if (current < 0) { // clicou fora de um quadrado existente => crie um
      //System.out.println(inicia+" "+inicia.ck_objeto);
      if (tipo == RETANGULO)
         addRetangulo(x, y);
      }
   else 
   if (evt.getClickCount() == 1) {
      ck = inicia.ck_objeto.getSelectedCheckbox(); //getCurrent();
      if (ck.getLabel().equals(S_ARESTA) ){
         moveObjeto = false;
         return;
         }
      moveObjeto = !moveObjeto; //moveObjetos(x,y, resp[1]);
      moveObjetoI = resp[0];
      moveObjetoT = resp[1];
      return;
     }
   else //  if (evt.getClickCount() >= 2) // clique duplo => remove objeto
      if (resp[1]==0) {
          //  System.out.println("narestas="+narestas);
          removeRetangulo(current);
          //  System.out.println("narestas="+narestas);
          }
   moveObjeto = false;
   }

 public void moveObjetos(int x, int y, int tipo){
   Graphics g = getGraphics();
   g.translate(getInsets().left, getInsets().top);
   g.setXORMode(getBackground()); // apaga ltimo desenho do quadrado atual - reescreve-o com cor de fundo
   if (tipo==0) {
      drawRetangulo(g, current); // apaga ltimo desenho
      for (int i = 0; i < MAXRETANGULOS; i++) // move arestas ligadas ao vrtice movido
          if (matAdj[current][i]>-1)
              drawAresta(g, arestas[matAdj[current][i]].indI,arestas[matAdj[current][i]].indF,
                            arestas[matAdj[current][i]].tipoI,arestas[matAdj[current][i]].tipoF);
      Retangulos[current].x = x; // atualiza coordenadas c/ nova posio
      Retangulos[current].y = y;
      drawRetangulo(g, current); // redesenha na nova posio
      for (int i = 0; i < MAXRETANGULOS; i++)
          if (matAdj[current][i]>-1)
              drawAresta(g, arestas[matAdj[current][i]].indI,arestas[matAdj[current][i]].indF,
                            arestas[matAdj[current][i]].tipoI,arestas[matAdj[current][i]].tipoF);
      }
   }


 public void mouseMoved(MouseEvent evt) {
   // Cursor.getPredefinedCursor( 1) = cursor tipo "+"                             CROSSHAIR_CURSOR
   // Cursor.getPredefinedCursor( 2) = cursor tipo "edio de texto"
   // Cursor.getPredefinedCursor( 3) = cursor tipo "relgio espera"
   // Cursor.getPredefinedCursor( 4) = cursor tipo "pega canto inferior esquerdo"
   // Cursor.getPredefinedCursor( 5) = cursor tipo "pega canto inferior direito"
   // Cursor.getPredefinedCursor( 6) = cursor tipo "pega canto superior esquerdo"
   // Cursor.getPredefinedCursor( 7) = cursor tipo "pega canto superior direito"
   // Cursor.getPredefinedCursor( 8) = cursor tipo "pega canto superior"
   // Cursor.getPredefinedCursor( 9) = cursor tipo "pega canto inferior"
   // Cursor.getPredefinedCursor(10) = cursor tipo "pega canto esquerdo"
   // Cursor.getPredefinedCursor(11) = cursor tipo "pega canto direito"
   // Cursor.getPredefinedCursor(12) = cursor tipo "dedo aponta"                   Cursor.HAND_CURSOR
   // Cursor.getPredefinedCursor(13) = cursor tipo "seta dupla em cruz"
   int x = evt.getX();
   int y = evt.getY();
   int [] resp = new int[2];
   resp = find(x, y); // pega o quadrado sob o mouse (se houver)
   // muda cursor se estiver sobre um objeto
   //moveObjetoI = resp[0];
   if (moveObjeto) {
       //System.out.println("move");
       current = moveObjetoI;
       moveObjetos(x,y, moveObjetoT); //
       }
   else
   if (resp[0] >= 0)
       this.setCursor(Cursor.getPredefinedCursor(12)); //(new Cursor(Cursor.HAND_CURSOR) ); //
   else
       this.setCursor(Cursor.getDefaultCursor());
   }
 
 public void mouseDragged(MouseEvent mouseevent) {
   /*
   int x = evt.getX();
   int y = evt.getY();
   int [] resp = new int[2];
   resp = find(x, y); // pega o quadrado sob o mouse (se houver)
   current = resp[0];  
   if (current >= 0) moveObjetos(x,y, resp[1]); */
   }

 public void drawRetangulo(Graphics g, int i) {
   int x = Retangulos[i].x,
       y = Retangulos[i].y;

   g.setColor( corBorda );
   g.drawRect( x - LARGURA_RETANGULO / 2-1, y - ALTURA_RETANGULO / 2-1,
               LARGURA_RETANGULO+1, ALTURA_RETANGULO+1);
   g.drawString( String.valueOf(rotRetangulos[i]),x+LARGURA_RETANGULO/2,y ); // x+RAIO/2,y );

   g.setColor( corRetangulo );
   g.fillRect( x - LARGURA_RETANGULO / 2, y - ALTURA_RETANGULO / 2,
               LARGURA_RETANGULO, ALTURA_RETANGULO);
   }


 // verifica se vrtice tipo t, com ndice i tem aresta
 public boolean buscaAdj(int vert, int t) {            // forma da matriz                     
   int i = vert,j,k;                                   //         <- quadrados ->|
   if (t==RETANGULO) {                                 //        |               :           |
       for (j = 0; j < MAXRETANGULOS; j++)             // quad.  |               :           |
            if (matAdj[i][j]>-1) return true;          //        |...............:...........|
       }                                               //
   return false;
   }

 public int [] coordAresta( int i, int f, int ti, int tf) {
   int [] resp = new int[4];

   int xi,yi,
       xf,yf;
   //System.out.println(" - desenha - i="+i+" f="+f);

   // Vrtice inicial
   if (ti == RETANGULO) {
      xi = Retangulos[i].x;
      yi = Retangulos[i].y; }
   else {
       System.out.println("[AreaDesenho!drawAresta]: erro, tipo inexistente");
       return null;
       }
   // Vrtice final
   if (tf == RETANGULO) {
      xf = Retangulos[f].x;
      yf = Retangulos[f].y; }
   else {
       System.out.println("[AreaDesenho!drawAresta]: erro, tipo inexistente");
       return null;
       }
   resp[0] = xi; resp[1] = yi;
   resp[2] = xf; resp[3] = yf;
   return resp;
   }


 public void drawAresta(Graphics g, int i, int f, int ti, int tf) {
   int [] resp = coordAresta(i,f,ti,tf);
   g.setColor(corAresta);
   g.drawLine(resp[0],resp[1],resp[2],resp[3]);
   }


 public int [] find(int x, int y) {
   int [] resp = new int[2];
   // verifica se existe quadrado sob mouse
   for (int i = 0; i < nquadrados; i++) // RAIO <-> LARGURA_RETANGULO
       if (Retangulos[i].x - LARGURA_RETANGULO / 2 <= x && x <= Retangulos[i].x + LARGURA_RETANGULO / 2 &&
           Retangulos[i].y - LARGURA_RETANGULO / 2 <= y && y <= Retangulos[i].y + LARGURA_RETANGULO / 2) {
           resp[0]=i; resp[1]=RETANGULO;
           return resp;
          }
   resp[0]=-1; resp[1]=-1;
   return resp;
   }


 public void addRetangulo(int x, int y) {
   if (nquadrados<0) System.out.println("nquadrados="+nquadrados);
   if (nquadrados < MAXRETANGULOS) {
      Retangulos[nquadrados] = new Point(x, y);
      rotRetangulos[nquadrados] = nrotquadrados;

      //Soluo 1: repaint
      //repaint(); // se usar o "repaint" ele redesenha a tela toda

      //Soluo 2:
      Graphics g = getGraphics();
      g.translate(getInsets().left, getInsets().top);
      g.setXORMode(getBackground()); // apaga ltimo desenho do quadrado atual - reescreve-o com cor de fundo
      drawRetangulo(g, nquadrados); // redesenha na nova posio

      nquadrados++; nrotquadrados++;
      }
   else 
      inicia.msg("Lamento, foi atingido o limite de vrtices tipo quadrado ("+MAXRETANGULOS+")");
 }
 
 public void removeRetangulo(int n) {
   int j;
   if (nquadrados>0 && n>-1) {
      nquadrados--;
      Retangulos[n]    = Retangulos[nquadrados];
      rotRetangulos[n] = rotRetangulos[nquadrados];
      if (current == n) current = -1;
      matAdj[n][n] = -1; // estava sujando diagonal principal
      //if ( buscaAdj(n,RETANGULO) ) { // tem arestas
         for (j = 0; j < MAXRETANGULOS; j++)
 	    if (matAdj[n][j]>-1) {
		//System.out.println("remove: "+matAdj[n][j]+" ("+n+","+j+")");
                removeAresta(matAdj[n][j]);
	        }
      //}

      repaint();//pinta();//
      }
   else System.out.println("nquadrados="+nquadrados+ " n=" +n);
 }
 


 // Encontra os ndices reais na matriz de adjacncia
 // til se forem introduzidos outros tipos de ns (bolas, ...) em posio subsequnte (aps MAXRETANGULOS)
 private int[] pegaIndice( int i, int f, int ti, int tf) {
   int [] resp = new int[2];
   if (ti==RETANGULO) { // insere arestas de nm. "narestas" na matriz de adjacncias
       if (tf==RETANGULO)
	  { resp[0] = i; resp[1] = f; }
       else
	  { resp[0] = i; resp[1] = MAXRETANGULOS+f; }
       }
   else {
       if (tf==RETANGULO)
	  { resp[0] = MAXRETANGULOS+i; resp[1] = f; }
       else
	  { resp[0] = MAXRETANGULOS+i; resp[1] = MAXRETANGULOS+f; }
       }
   return resp;
 }


 // nd. incio aresta, nd. fim aresta, tipo incio aresta, tipo fim aresta
 public void insereAresta( int i, int f, int ti, int tf) {
   int [] resp = pegaIndice( i, f, ti, tf);
   int i1 = resp[0], f1 = resp[1];
   if ( matAdj[i1][f1] > -1 ) {// ||matAdj[i][f] > -1)
       //System.out.println("aresta j existente: ("+i+","+f+")");
       //inicia.p_inf.msg.setText("aresta j existente: ("+i+","+f+")");
       inicia.msg("aresta j existente: ("+i+","+f+")");
       return;
       }
   if (narestas>=MAXRETANGULOS) {
       System.out.println("[AreaDesenho!insereAresta]: erro, excesso de arestas");
       return;
       }
   matAdj[i1][f1] = matAdj[f1][i1] = narestas;
   arestas[narestas] = new Arestas(i, f, ti, tf);
   narestas++;
   Graphics g = getGraphics();
   drawAresta(g, i, f, ti, tf);
   }
 
// remove aresta ligando vrtice i com vrtice j
 public void removeAresta(int n) {
   int i = arestas[n].indI,
       f =arestas[n].indF;

   if (narestas>0 && n>-1) {
      narestas--;
      //System.out.println("Remove: "+n+" inicio="+arestas[n].indI+" fim="+arestas[n].indF);
      if (arestas[n].tipoI == RETANGULO) {
          if (arestas[n].tipoF == RETANGULO)
              matAdj[i][f] = matAdj[f][i] = -1;
	 else 
              matAdj[i][MAXRETANGULOS+f] = matAdj[MAXRETANGULOS+f][i] = -1;
          }
      else {
          if (arestas[n].tipoF == RETANGULO)
              matAdj[MAXRETANGULOS+i][f] = matAdj[f][MAXRETANGULOS+i] = -1;
	 else 
              matAdj[MAXRETANGULOS+i][MAXRETANGULOS+f] = matAdj[MAXRETANGULOS+f][MAXRETANGULOS+i] = -1;
          }
      arestas[n] = arestas[narestas];
      // if (current == n) current = -1;
      repaint();//pinta();//
      }
   else System.out.println("narestas="+narestas+ " n=" +n);
 }


 // Chamado sempre que precisa atualizar desenho
 public void pinta() {
   Graphics g = offscreen.getGraphics();//null;  
   //offscreen = null; // para apagar ltimo desenho, caso contrrio fica na tela
   // g.translate(getInsets().left, getInsets().top);
   // g.setXORMode(getBackground()); // apaga ltimo desenho do quadrado atual - reescreve-o com cor de fundo

   paint(g); 
 }

 private void Retardo( ) {
   try {
    Thread.sleep(steps * DELAY);
   } catch (InterruptedException ie) { }
 }

 public synchronized void paint(Graphics g) { // synchronized 
   Dimension d=this.getSize(); //size();

   setBackground(Color.white);
   
   if (offscreen == null) // se offscreen vazio, inicie-o
       offscreen = this.createImage(d.width, d.height); //
   
   //  g = offscreen.getGraphics(); // est ruim com este buffer

   g.translate(getInsets().left, getInsets().top);
   for (int i = 0; i < nquadrados; i++) { // desenha quadrados
       //g.setColor( corRetangulo );
       drawRetangulo(g, i);
       }
   for (int i = 0; i < narestas; i++) {   // 
       int [] resp = coordAresta(arestas[i].indI, arestas[i].indF, arestas[i].tipoI, arestas[i].tipoF);
       g.setColor(corAresta);
       g.drawLine(resp[0],resp[1],resp[2],resp[3]);
       }

   g.setColor(Color.black);
   g.drawRect( 0,0, this.getSize().width-1,this.getSize().height-1);
   g.drawRect( 3,3, this.getSize().width-7,this.getSize().height-7);

   // deprecated
   // g.drawRect( 0,0, this.size().width-1,this.size().height-1);
   // g.drawRect( 3,3, this.size().width-7,this.size().height-7);

   //g.drawLine(0,0,  this.size().width,this.size().height);
 
   // copia tudo na tela
   g = this.getGraphics();
   //  g.drawImage(offscreen,0,0,this); // est ruim com este buffer
   } 

}
